"""A representation of the packages a Godot project has installed."""

from pathlib import Path
from gopm import git
from typing import List, Tuple
from send2trash import send2trash
import random
import shutil

class Package:
    """An installable Godot package."""

    def __init__(self, uri: str, version: str):
        "The Git URI the package can be cloned from."
        self.uri = uri
        "The commit hash to check out."
        self.version = version
        self.name = Path(uri).stem

    def get_latest_version(self, tmp: Path) -> str:
       """Clones the repositories of the package and returns the
       latest version available.
       """
       repo = tmp / self.name
       git.clone_repo(self.uri, repo)
       latest = git.get_latest_commit(repo)[:8]
       shutil.rmtree(repo)
       return latest


class Project:
    """A Godot project with a list of installed packages."""

    def __init__(self, path: Path | str):
        self.path = Path(path)
        self.modules_file = Path(path) / "godotmodules.txt"

    def get_installed(self) -> List[Package]:
        """Reads the modules file and lists the installed packages."""
        try:
            with open(self.modules_file) as file:
                return list(map(lambda x: Package(*x.split()),
                        filter(lambda x: x.strip() != "", file)))
        except FileNotFoundError:
            return []

    def download_addons(self, package: Package, tmp: Path | str
            ) -> Tuple[List[str], List[Package]]:
        """Install the addons of a package.
        
        Clones the repository of the package and copies the addons
        inside the addons folder into the project. Returns the list
        of installed addons and the dependencies of the package.
        `tmp` is the folder into which the repository will be cloned.
        
        Existing addons with the same name as installed addons will be
        moved to the trash.

        *Example:*

        ```python
        (addons, dependencies) = project.download_addons(package, Path())
        ```
        """
        repo = Path(tmp) / package.name
        git.clone_repo(package.uri, repo, package.version)
        addons = self.path / "addons"
        addons.mkdir(exist_ok=True)
        addon_names : List[str] = []
        for addon in (repo / "addons").glob("*"):
            addon_names.append(addon.stem)
            destination = addons / addon.name
            if destination.is_dir():
                send2trash(destination)
            shutil.copytree(addon, destination)
        project = Project(repo)
        sub_packages = project.get_installed()
        shutil.rmtree(repo)
        return (addon_names, sub_packages)


    def save_packages(self, packages: List[Package]):
        """Writes the package list to the godotmodules.txt file."""
        with open(self.modules_file, "w") as file:
            file.writelines(map(lambda x: f"{x.uri} {x.version}\n", packages))



"""Godot Package Manager.

This module can be used as a standalone command-line tool, or accessed with the
Python API.

## Examples

Install an addon:

```python
from gopm import Project
import tempfile

tmp = Path(tempfile.gettempdir()) / "godot_packages"
# Project at current working directory.
project = Project(Path())
(addons, dependencies) = project.download_addons(package, tmp)
print("Installed addons: ", addons)
print("Dependencies: ", dependencies)
```

Update addons:

```python
from gopm import Project
import tempfile

tmp = Path(tempfile.gettempdir()) / "godot_packages"
project = Project(Path())
for dep in project.get_installed():
    project.download_addons(dep, tmp)
```

Get a list of addons:

```python
from gopm import Project

installed = Project("../tetris").get_installed()
print("\n".join(map(lambda p: p.name, installed))
```
"""

from gopm.project import Project
from gopm.project import Package


"""Git utilities."""

import subprocess
import urllib
from typing import List
from pathlib import Path

def clone_repo(uri: str, to: Path, version: str | None = None):
    """Clone the given repository to the given path.

    When passed, version will be switched to in a detached HEAD state.
    """
    _run_git_command(["clone", uri, str(to.absolute())])
    if version is not None:
        _run_git_command(["switch", "--detach", version], to)


def get_latest_commit(repo: Path) -> str:
    """Return the hash of the current commit of the main branch."""
    return _run_git_command(["rev-parse", "HEAD"], repo).strip()


def _run_git_command(args: List[str], working_directory = Path()) -> str:
    """Run a Git command inside working_directory.

    Returns the content of stdout after running the command.
    """
    args.insert(0, "git")
    result = subprocess.run(args, capture_output=True, cwd=working_directory,
            check=True)
    return result.stdout.decode('UTF-8')


"""The main command line interface of the package manager."""

import shutil
import sys
import tempfile

from argparse import ArgumentParser
from gopm import git
from gopm import Project, Package
from gopm.search.github import GithubSearchProvider
from gopm.search.search_provider import Result, SearchProvider
from pathlib import Path
from subprocess import CalledProcessError
from typing import List, Tuple

tmp_repos_dir = Path(tempfile.gettempdir()) / "godot_packages"
search_providers = [GithubSearchProvider()]

def show_install_help():
    """Print information on how to use the utility."""

    print("""No packages installed. Install one like this:

    > gopm install [search term]

    or

    > gopm install /path/to/git/repo.git
    """)


def download_addons(project: Project, package: Package, indent: int = 0):
    """Download addons of a project and their dependencies."""
    indent_str = '\t' * indent
    version = package.version[:8]
    print(f"{indent_str}[{package.name}] version {version} from {package.uri}")
    (addons, depedencies) = project.download_addons(package, tmp_repos_dir)
    for addon in addons:
        print(f"\t{indent_str}Addon [{addon}]")
    if len(addons) == 0:
        print(f"\t{indent_str}Package contains no addons.")
    for dependency in depedencies:
        addons += download_addons(project, dependency, indent + 1)
    return addons


def install(project: Project, search: List[str]):
    """Install a package from a path or select from a list of
    repositories matching the search.
    """
    term = " ".join(search)
    installed = project.get_installed()
    path = Path(term)
    package : Package
    if path.is_dir():
        package = Package(term, "")
        package.version = package.get_latest_version(tmp_repos_dir)
    else:
        results : List[Result] = []
        for provider in search_providers:
            results += provider.get_matches(term)
        if len(results) == 0:
            return print("No packages found")
        print("\n".join(map(
                lambda x: f"[{x[0] + 1}] {x[1].name}\n\t{x[1].description}",
                enumerate(results))))
        try:
            package_num = int(input("Package to install: "))
        except (ValueError, KeyboardInterrupt):
            return print("No package selected")
        selected = results[package_num - 1]
        package = Package(selected.url, selected.latest_version)
    if any(map(lambda x: x.name == package.name, installed)):
        return print("Package is already installed.")
    project.save_packages(installed + [package])
    installed = download_addons(project, package)

    ignore_path = project.path / ".gitignore"
    MARKER = "# Godot Packages\n"
    if ignore_path.is_file():
        with open(ignore_path) as ignore:
            ignored = ignore.readlines()
        if MARKER not in ignored:
            ignored += ["\n", MARKER]
        marker_pos = ignored.index(MARKER)
        for addon in installed:
            ignore_str = f"addons/{addon}\n"
            if ignore_str not in ignored:
                ignored.insert(marker_pos + 1, ignore_str)
        with open(ignore_path, "w") as ignore:
            ignore.writelines(ignored)

    print(f"Installed {package.name}")


def update(project: Project):
    """Download the addons of the installed packages."""
    if len(project.get_installed()) == 0:
        show_install_help()
    else:
        for package in project.get_installed():
            download_addons(project, package)


def upgrade(project: Project):
    """Update the addons of the installed packages to the latest
    version.
    """
    if len(project.get_installed()) == 0:
        return show_install_help()
    packages = project.get_installed()
    for package in packages:
        latest = package.get_latest_version(tmp_repos_dir)
        if package.version == latest:
            print(f"{package.name} is up-to-date.")
            continue
        print(f"Upgrading {package.name} from {package.version} to {latest}")
        package.version = latest
        download_addons(project, package)
    project.save_packages(packages)


def remove(project: Project, package: str):
    """Remove a package that contains a query."""
    installed = project.get_installed()
    matches = list(filter(lambda p: package.lower() in p.name.lower(),
            installed))
    match len(matches):
        case 0:
            print(f'No package installed that contains the word "{package}"')
        case 1:
            match = matches[0]
            answer = input(f'Really remove "{match.name}"?\n[y/N] ')
            if answer and answer.lower() in {"y", "yes"}:
                installed.remove(match)
                project.save_packages(installed)
                print(f"Removed {match.name}")
        case _:
            names = ', '.join(map(lambda p: p.name, matches))
            print(f"Found multiple results, be more specific: {names}")


def list_packages(project: Project):
    """Print a list of installed packages."""
    installed = project.get_installed()
    if len(installed) == 0:
        show_install_help()
    for package in installed:
        print(f"{package.name} [{package.version}]")


def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()
    update_parser = subparsers.add_parser("update", aliases=["u"],
            help=("Download all packages inside the modules file and installs" +
                 " their addons. WARNING: Existing addons with the same name" +
                 " will be moved to the trash."))
    update_parser.set_defaults(func=update)
    upgrade_parser = subparsers.add_parser("upgrade", aliases=["s"],
            help="Update all packages to the latest version")
    upgrade_parser.set_defaults(func=upgrade)
    list_parser = subparsers.add_parser("list", aliases=["l"],
            help="List the installed packages")
    list_parser.set_defaults(func=list_packages)
    install_parser = subparsers.add_parser("install", aliases=["i"],
            help="Install a package from a git URI or search and install a\
            package from Github")
    install_parser.add_argument("search",
            help="Git URI or name to search on Github", nargs="+")
    install_parser.set_defaults(func=install)
    remove_parser = subparsers.add_parser("remove", aliases=["r"],
            help="Uninstall the specified package")
    remove_parser.add_argument("package",
            help="Name to match in the modules file")
    remove_parser.set_defaults(func=remove)

    args = vars(parser.parse_args())
    try:
        func = args.pop("func")
    except KeyError:
        return parser.print_help()

    if tmp_repos_dir.exists():
        shutil.rmtree(tmp_repos_dir)
    tmp_repos_dir.mkdir()
    try:
        func(Project(Path()), **args)
    except CalledProcessError as e:
        err = e.stderr.decode("unicode_escape").strip()
        cmd = " ".join(e.cmd)
        print(f'Error running Git command "{cmd}":\n\t{err}')
    shutil.rmtree(tmp_repos_dir)


if __name__ == "__main__":
    main()



"""Classes for searching package databases."""

from typing import List

class Result:
    """A single result from a search."""

    def __init__(self, name: str, description: str, latest_version: str,
            url: str):
        self.name = name
        "The name of the repo."
        self.description = description
        self.url = url
        "The Git URI to clone the repo."
        self.latest_version = latest_version
        "The commit hash of the latest commit."

    @property
    def latest_version(self) -> str:
        return self._latest_version

    @latest_version.setter
    def latest_version(self, value) -> None:
        self._latest_version = value

class SearchProvider:
    """A singleton which allows searching a forge for Godot addons."""

    def __init__(self):
        raise NotImplementedError

    def get_matches(self, term: str) -> List[Result]:
        """Make a web request to retrieve a list of addons matching a
        search query.
        """
        raise NotImplementedError




"""Search provider using Github's API."""

import json
import requests

from gopm.search.search_provider import SearchProvider, Result
from typing import List

class GithubResult(Result):
    def __init__(self, content: dict):
        super().__init__(content["full_name"], content["description"],
                "master", content["clone_url"])
        self.commits_url = content["commits_url"].replace('{/sha}', "")

    @property
    def latest_version(self) -> str:
        return requests.get(self.commits_url).json()[0].get("sha")[:8]

    @latest_version.setter
    def latest_version(self, value) -> None:
        self._latest_version = value


class GithubSearchProvider(SearchProvider):
    def __init__(self):
        pass
    
    def get_matches(self, term: str) -> List[Result]:
        query = f"https://api.github.com/search/repositories?q=\
    {term} language:GDScript&per_page=10"
        text = requests.get(query).text
        items = json.loads(text).get("items")
        return list(map(GithubResult, items))
